/* -*- c-basic-offset: 4 -*- */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdarg.h>
#include <errno.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "net.h"
#include "server.h"

#define MAX_CLIENTS 10
#define SRV_FDS 1

struct client {
    int fd;
    char buf[BUFSIZ];
    int bufend;
    int linetoolong;
};

int servsock;
struct client clients[MAX_CLIENTS];
struct pollfd pollfds[SRV_FDS+MAX_CLIENTS];
int nclients = 0;

int
errx(int eval, const char *fmt, ...) {
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(eval);
}

void
parse_linetoolong(int clientid, char *line) {
    printf("linetoolong: %d \"%s\"\n", clientid, line);
}

void
net_server(int port) {
    struct sockaddr_in serveraddr;
    serveraddr.sin_family = AF_INET;
    serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
    serveraddr.sin_port = htons(port);


    servsock=socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (servsock<0)
	errx(1, "socket: %s\n", strerror(errno));
    int r;
    port--;
    do {
	serveraddr.sin_port = htons(++port);
	r=bind(servsock, (struct sockaddr *)&serveraddr, sizeof(serveraddr));
    } while ((r<0) && (errno==EADDRINUSE));
    if (r < 0)
	errx(1, "bind port %d: %s\n", port, strerror(errno));
    if (listen(servsock, 10) < 0)
	errx(1, "listen: %s\n", strerror(errno));
    pollfds[0].fd=servsock;
    pollfds[0].events=POLLIN|POLLERR|POLLHUP;
    for (int i=0; i<MAX_CLIENTS; i++) {
	pollfds[SRV_FDS+i].fd=-1;
	pollfds[SRV_FDS+i].events=0;
    }
    printf("Waiting for connections on port %d...\n", port);
}


void
net_all_flush() {
}

void
net_client_printf(int clientid, const char *fmt, ...) {
    va_list ap;
    char buf[BUFSIZ];
    int len, n;
    struct client *c = &clients[clientid];
    
    va_start(ap, fmt);
    len=vsnprintf(buf, sizeof(buf)-1, fmt, ap);
    va_end(ap);
    if (len > BUFSIZ-1)
	errx(1, "BUFSIZ too small\n");
    char *bufp=buf;
    do {
	n=write(c->fd, bufp, len);
	if (n<0)
	    errx(1, "write: %s\n", strerror(errno));
	bufp+=n;
    } while (bufp < buf+len);
}

void
net_all_printf(const char *fmt, ...) {
    va_list ap;
    char buf[BUFSIZ];
    int len, n;
    va_start(ap, fmt);
    len=vsnprintf(buf, sizeof(buf)-1, fmt, ap);
    va_end(ap);
    if (len > BUFSIZ-1)
	errx(1, "BUFSIZ too small\n");
    for (int clientid=0; clientid<2; clientid++) {
	struct client *c = &clients[clientid];
	char *bufp=buf;
	do {
	    n=write(c->fd, bufp, len);
	    if (n<0)
		errx(1, "write: %s\n", strerror(errno));
	    bufp+=n;
	} while (bufp < buf+len);
    }
}


void net_client_accept(int servsock);
void net_client_read(int clientid);

void
net_poll() {
    int npoll;
    npoll=poll(pollfds, SRV_FDS+MAX_CLIENTS, -1);
    if (npoll<0)
	errx(1, "poll: %s", strerror(errno));
    if (pollfds[0].revents) {
	net_client_accept(pollfds[0].fd);
	npoll--;
    }
    for (int i=0; npoll && i<MAX_CLIENTS; i++) {
	if (pollfds[SRV_FDS+i].revents) {
	    net_client_read(i);
	    npoll--;
	}
    }
}


void
net_client_accept(int servsock) {
    struct sockaddr_in clientaddr;
    int cfd;
    socklen_t slen;
    cfd=accept(servsock, (struct sockaddr *)&clientaddr, &slen);
    if (cfd<0)
	errx(1, "accept: %s", strerror(errno));
    // FIXME log, shout, or something
    if (nclients>MAX_CLIENTS) {
#define STR_SERVER_FULL "Sorry, the server is full\n"
	write(cfd, STR_SERVER_FULL, sizeof(STR_SERVER_FULL)-1);
	close(cfd);
    } else {
	int clientid;
	for (clientid=0; clientid<MAX_CLIENTS; clientid++)
	    if (pollfds[SRV_FDS+clientid].fd == -1) break;
	
	printf("Client %d connected...\n", clientid);
	pollfds[SRV_FDS+clientid].fd=cfd;
	pollfds[SRV_FDS+clientid].events=POLLIN|POLLERR|POLLHUP;
	
	clients[clientid].fd = cfd;
	clients[clientid].linetoolong = 0;
	clients[clientid].bufend = 0;

	server_client_connect(clientid);
			      
	nclients++;
    }
}

void
net_client_close(int clientid) {
    printf("Closing down client %d...\n", clientid);

    close(pollfds[SRV_FDS+clientid].fd);

    pollfds[SRV_FDS+clientid].revents = 0;
    pollfds[SRV_FDS+clientid].fd = -1;

    clients[clientid].fd = -1;
    clients[clientid].linetoolong = 0;
    clients[clientid].bufend = 0;

    nclients--;
}

void
net_client_read(int clientid) {
    struct client *c = &clients[clientid];
    int len = c->bufend;
    int n;
    
    n = read(c->fd, c->buf + len, BUFSIZ - len);
    // printf("read client %d %d\n", clientid, n);
    if (n==0) {
	net_client_close(clientid);
	return;
    }
    if (n<0)
	errx(1, "read fail: %s\n", strerror(errno));

    char *line = c->buf;
    char *bufend = c->buf + n;
    char *p = c->buf + len; /* start search for '\n' at end of last buffer */
    while ((p = memchr(p, '\n', bufend - p))) {
	*p='\0';
	if ((p - line > 0) && (*(p-1) == '\r'))
	    *(p-1)='\0';
	if (c->linetoolong) {
	    c->linetoolong = 0;
	} else {
	    parse_line(clientid, line);
	}
	line=++p; // start of unparsed buffer
    }
    if (line == c->buf) {
	/* linestart at start of buffer ie nothing was parsed */
	if ((bufend == c->buf + BUFSIZ) && !c->linetoolong) {
	    /* buffer is full with overlong line
	     * send to parser for error msg or partial parsing */
	    c->linetoolong = 1;
	    c->buf[BUFSIZ-1]='\0';
	    parse_linetoolong(clientid, line);
	}
	if (c->linetoolong)
	    c->bufend = 0;
    } else {
	if (bufend - line) {
	    /* move partial line to beginning of buffer */
	    memmove(c->buf, line, bufend - line);
	    c->bufend = bufend - line;
	}
    }
}

